import tkinter as tk
import traceback
from logging import getLogger
from tkinter import messagebox
from typing import List, Optional

from turcar import editor_helpers, get_runner, get_workbench
from turcar.codeview import SyntaxText
from turcar.common import InlineCommand, SignatureInfo
from turcar.editor_helpers import DocuBoxBase, get_active_text_widget, get_text_filename
from turcar.languages import tr
from turcar.shell import ShellText
from turcar.ui_utils import ems_to_pixels

logger = getLogger(__name__)


class CalltipBox(DocuBoxBase):
    def __init__(self, calltipper: "Calltipper"):
        super(CalltipBox, self).__init__(show_vertical_scrollbar=False)
        self._calltipper = calltipper
        self._max_width = 40

    def present_signatures(self, text: SyntaxText, signatures: List[SignatureInfo]):
        self._target_text_widget = text
        self._check_bind_for_keypress(text)
        self.text.direct_delete("1.0", "end")
        self.render_signatures(signatures, only_params=True)

        char_count = self.text.count("1.0", "end", "chars")[0]
        extra_width_factor = 1.0
        # NB! Width should be set here, not in update_size, so that it that on-screen
        # height calculation already uses this
        if char_count * extra_width_factor < self._max_width:
            self.text["width"] = max(round(char_count * extra_width_factor), 10)
        else:
            self.text["width"] = self._max_width

        self.text["height"] = 3

        expected_height = 10  # TODO
        pos = signatures[0].call_bracket_start
        if pos:
            self._show_on_target_text(
                "%d.%d" % pos,
                expected_height,
                "above",
                y_offset=-ems_to_pixels(0.3),
            )

    def _update_size(self):
        # will be called while box is off the screen
        display_line_count = self.text.count("1.0", "end", "displaylines")[0]
        self.text["height"] = display_line_count

    def _on_text_keypress(self, event=None):
        if not self.is_visible():
            return

        if event.keysym == "Escape":
            return

        if event.char or event.keysym in [
            "Left",
            "Right",
            "Up",
            "Down",
            "KP_Left",
            "KP_Right",
            "KP_Up",
            "KP_Down",
        ]:
            self.after_idle(self._calltipper.request_calltip)


class Calltipper:
    def __init__(self):
        self._calltip_box: Optional[CalltipBox] = None
        get_workbench().bind("get_editor_calltip_response", self.handle_response, True)
        get_workbench().bind("get_shell_calltip_response", self.handle_response, True)
        get_workbench().bind("AutocompletionInserted", self._on_autocomplete_insertion, True)
        get_workbench().bind_class("EditorCodeViewText", "<Key>", self._on_text_key, True)
        get_workbench().bind_class("ShellText", "<Key>", self._on_text_key, True)

    def request_calltip(self) -> None:
        text = get_active_text_widget()
        if not text:
            return
        self.request_calltip_for_text(text)

    def request_calltip_for_text(self, text: SyntaxText) -> None:
        source, row, column = editor_helpers.get_relevant_source_and_cursor_position(text)
        get_runner().send_command(
            InlineCommand(
                "get_shell_calltip" if isinstance(text, ShellText) else "get_editor_calltip",
                row=row,
                column=column,
                source=source,
                filename=get_text_filename(text),
            )
        )

    def handle_response(self, msg) -> None:
        text = get_active_text_widget()
        if not text:
            return

        source, row, column = editor_helpers.get_relevant_source_and_cursor_position(text)

        if msg.get("error"):
            self._hide_box()
            messagebox.showerror("Calltip error", msg.error, master=get_workbench())
        elif msg.source != source or msg.row != row or msg.column != column:
            # situation has changed, information is obsolete
            return
        elif not msg.signatures:
            logger.debug("Back-end gave 0 signatures")
            self._hide_box()
        else:
            if not self._calltip_box:
                self._calltip_box = CalltipBox(self)
            self._calltip_box.present_signatures(text, msg.signatures)

    def _on_autocomplete_insertion(self, event=None):
        text = get_active_text_widget()
        if not text:
            return

        if (
            self._box_is_visible()
            or self._should_show_automatically()
            and text.get("insert -1 chars") == "("
        ):
            self.request_calltip_for_text(text)

    def _should_show_automatically(self) -> bool:
        return get_workbench().get_option("edit.automatic_calltips")

    def _box_is_visible(self):
        return self._calltip_box and self._calltip_box.is_visible()

    def _hide_box(self) -> None:
        if self._calltip_box:
            self._calltip_box.hide()

    def _on_text_key(self, event: tk.Event) -> None:
        if event.char == "(" and self._should_show_automatically():
            self.request_calltip_for_text(event.widget)


def load_plugin():
    calltipper = Calltipper()

    get_workbench().set_default("edit.automatic_calltips", False)

    get_workbench().add_command(
        "open_calltip",
        "edit",
        tr("Show parameter info"),
        calltipper.request_calltip,
        # TODO: tester
        default_sequence="<<ShiftControlSpaceInText>>",
        accelerator="Ctrl-Shift-Space",
    )



import itertools


def print_combinations(people):
    result = []  # 用于存储所有可能的过河方案
    left_bank = set(people)  # 左岸的初始情况
    right_bank = set()  # 右岸的初始情况
    boat = set()  # 船上的初始情况

    # 判断双人组是否安全
    def is_valid(path):
        if ('妈妈' in path and ('爸爸' in path or '路人' in path or '哥哥' in path)) or (
                '爸爸' in path and ('妹妹' in path or '妈妈' in path)) or (
                '哥哥' in path and ('妹妹' in path or '妈妈' in path or '路人' in path)) or (
                '妹妹' in path and ('爸爸' in path or '哥哥' in path)) or (
                '路人' in path and ('妈妈' in path or '哥哥' in path)):
            return False
        return True

    # 模拟每次过河的情况
    def cross(left_river, path=[]):
        print("左： ", left_river)
        print("右： ", right_bank)

        # 返回左岸,从船上下来一个人
        if boat:
            # print("回来的人： ", boat)
            left_river.update(boat)  # 船上的人回到左岸
            right_bank.difference_update(boat)  # 船上的人从右岸移除
            # print("path： ", path)
            path = path + [(list(boat)[0])]  # 过河方案拼接
            boat.clear()
            # print("path： ", path)
            # print("回来后左岸： ", left_river)

        # 遍历所有可能的过桥方式，即从左岸选取两个人
        for pair in itertools.combinations(left_river, 2):
            a, b = pair
            # 安全的双人组
            if is_valid([a, b]):
                new_river = left_river - {a, b}  # 更新左岸的情况

                right_bank.update({a, b})  # 更新右岸的情况
                # print("过河后左岸:", new_river)
                # print("过河后右岸:", right_bank)
                new_path = path + [(a, b)]  # 过河方案拼接
                print("过河组合: ", new_path)

                # 如果左岸和船上都没有人，则表示所有人都已经成功过河，将当前路径加入结果中
                if not new_river and not boat:
                    result.append(path)
                    return
                # 遍历所有可能返回方式，
                for load in itertools.combinations(right_bank, 1):
                    if len(load) == 1 and load[0] != "妹妹":
                        # 执行相关操作
                        boat.clear()
                        boat.update(load)
                        cross(new_river, new_path)

    cross(left_bank)
    print("结束")
    print(result)
    for i, step in enumerate(result):
        print(f"Step {i + 1}: {step}")


people = ['妈妈', '爸爸', '哥哥', '妹妹', '路人']
print_combinations(people)



"""
Module: '_api.motionsensor' on micropython-v1.14-893-lego learning system hub
"""
# MCU: {'machine': 'LEGO Technic Large Hub with STM32F413xx', 'sysname': 'LEGO Technic Large Hub', 'platform': 'LEGO Learning System Hub', 'nodename': 'LEGO Learning System Hub', 'ver': 'v1.14-893', 'release': '1.14.0', 'name': 'micropython', 'family': 'micropython', 'port': 'LEGO Learning System Hub', 'version': '1.14.0', 'mpy': 517, 'build': '893'}
# Stubber: 1.7.2
from typing import Any


class MotionSensor():
    def __init__(self, *argv, **kwargs) -> None:
        """"""

    BACK = 'back' # type: str
    DOUBLE_TAPPED = 'doubletapped' # type: str
    DOWN = 'down' # type: str
    FALLING = 'falling' # type: str
    FRONT = 'front' # type: str
    LEFT_SIDE = 'leftside' # type: str
    RIGHT_SIDE = 'rightside' # type: str
    SHAKEN = 'shaken' # type: str
    TAPPED = 'tapped' # type: str
    UP = 'up' # type: str
    def align_to_model(self, *args, **kwargs) -> Any:
        """"""

    def get_gesture(self, *args, **kwargs) -> Any:
        """"""

    def get_orientation(self, *args, **kwargs) -> Any:
        """"""

    def get_pitch_angle(self, *args, **kwargs) -> Any:
        """"""

    def get_roll_angle(self, *args, **kwargs) -> Any:
        """"""

    def get_yaw_angle(self, *args, **kwargs) -> Any:
        """"""

    def reset_yaw_angle(self, *args, **kwargs) -> Any:
        """"""

    def wait_for_new_gesture(self, *args, **kwargs) -> Any:
        """"""

    def wait_for_new_orientation(self, *args, **kwargs) -> Any:
        """"""

    def was_gesture(self, *args, **kwargs) -> Any:
        """"""

def sleep_ms(*args, **kwargs) -> Any:
    """"""




from typing import Iterable, Tuple

from microbit import _MicroBitAnalogDigitalPin

class AudioFrame(list):
    """
    Represents a list of 32 samples each of which is a signed byte.
    It takes just over 4ms to play a single frame.
    """
    def copyfrom(self, other) -> None:
        pass



def play(
    source: Iterable[AudioFrame],
    wait: bool = True,
    pin: _MicroBitAnalogDigitalPin = ...,
    return_pin: _MicroBitAnalogDigitalPin = None
) -> None:
    """
    Play the source to completion.

    `source`: `Sound` - The microbit module contains a list of built-in sounds that your can pass to
    audio.play(). The source agrument can also be an iterable of AudioFrame elements as described below.

    If `wait` is True, this function will block until the source is exhausted.

    `pin` An optional argument to specify the output pin can be used to override the default of
    pin0. If we do not want any sound to play we can use pin=None.

    `return_pin`: specifies a differential edge connector pin to connect to an external speaker
    instead of ground. This is ignored for the V2 revision.
    """


def is_playing() -> bool:
    """
    Return True if audio is playing, otherwise return False.
    """


def stop() -> None:
    """
    Stops all audio playback.
    """



from pybricks.parameters import Port


class AnalogSensor:
    """
    Generic or custom analog sensor.

    Args:
        port (Port): Port to which the sensor is connected.
    """

    def __init__(self, port: Port):
        if port == Port.A or port == Port.B or port == Port.C or port == Port.D:
            raise ValueError("Sensors must use Port S1, S2, S3, or S4.")

    def voltage(self) -> int:
        """
        Measures analog voltage.

        Returns:
            Analog voltage in millivolts.
        """
        return 0

    def resistance(self) -> int:
        """
        Measures resistance.

        This value is only meaningful if the analog device is a passive load such as a resistor or thermistor.

        Returns:
            Resistance of the analog device in ohms.
        """
        return 0

    def active(self):
        """
        Sets sensor to active mode. This sets pin 5 of the sensor port to high.

        This is used in some analog sensors to control a switch. For example, if you use the NXT Light Sensor as a custom analog sensor, this method will turn the light on. From then on, voltage() returns the raw reflected light value.
        """
        ...

    def passive(self):
        """
        Sets sensor to passive mode. This sets pin 5 of the sensor port to low.

        This is used in some analog sensors to control a switch. For example, if you use the NXT Light Sensor as a custom analog sensor, this method will turn the light off. From then on, voltage() returns the raw ambient light value.

        """
        ...



# -*- coding: utf-8 -*-
import _thread
import io
import os.path
import pathlib
import queue
import stat
import sys
import threading
import time
import traceback
import warnings
from abc import ABC, abstractmethod
from logging import getLogger
from typing import Any, BinaryIO, Callable, Dict, Iterable, List, Optional, Tuple, Union

import turcar
from turcar import report_time
from turcar.common import (  # TODO: try to get rid of this
    IGNORED_FILES_AND_DIRS,
    PROCESS_ACK,
    BackendEvent,
    CommandToBackend,
    EOFCommand,
    ImmediateCommand,
    InlineCommand,
    InlineResponse,
    InputSubmission,
    MessageFromBackend,
    ToplevelCommand,
    ToplevelResponse,
    UserError,
    execute_with_frontend_sys_path,
    is_local_path,
    parse_message,
    read_one_incoming_message_str,
    serialize_message,
    try_load_modules_with_frontend_sys_path,
    universal_dirname,
)

NEW_DIR_MODE = 0o755


logger = getLogger(__name__)


class BaseBackend(ABC):
    """Methods for both MainBackend and forwarding backend"""

    def __init__(self):
        self._current_command = None
        self._incoming_message_queue = queue.Queue()  # populated by the reader thread
        self._interrupt_lock = threading.Lock()
        self._last_progress_reporting_time = 0
        self._last_sent_output = ""
        self._init_command_reader()

    def _init_command_reader(self):
        # NB! This approach is used only in MicroPython and SshCPython backend.
        # MainCPython backend uses main thread for reading commands
        # https://github.com/turcar/turcar/issues/1363
        threading.Thread(target=self._read_incoming_messages, daemon=True).start()

    def mainloop(self):
        report_time("Beginning of mainloop")

        try:
            while True:
                self._check_for_connection_error()
                try:
                    msg = self._fetch_next_incoming_message(timeout=0.01)
                except KeyboardInterrupt:
                    self._send_output(
                        "\nKeyboardInterrupt", "stderr"
                    )  # CPython idle REPL does this
                    self.send_message(ToplevelResponse())
                except queue.Empty:
                    self._perform_idle_tasks()
                else:
                    if isinstance(msg, InputSubmission):
                        self._handle_user_input(msg)
                    elif isinstance(msg, EOFCommand):
                        self._handle_eof_command(msg)
                    else:
                        self._current_command = msg
                        self._handle_normal_command(msg)
        except KeyboardInterrupt:
            self._send_output("\nKeyboardInterrupt", "stderr")
            sys.exit(0)
        except ConnectionError as e:
            self.handle_connection_error(e)
        except Exception:
            # Error in turcar's code
            logger.exception("mainloop error")
            self._report_internal_exception("mainloop error")

        logger.info("After mainloop")
        sys.exit(17)

    def handle_connection_error(self, error=None):
        logger.info("Handling connection error")
        message = "Connection lost"
        if error:
            message += " -- " + str(error)
        self._send_output("\n" + message + "\n", "stderr")
        self._send_output("\n" + "Use Stop/Restart to reconnect." + "\n", "stderr")
        sys.exit(1)

    def _current_command_is_interrupted(self):
        return getattr(self._current_command, "interrupted", False)

    def _fetch_next_incoming_message(self, timeout=None):
        return self._incoming_message_queue.get(timeout=timeout)

    def _report_progress(
        self, cmd, description: Optional[str], value: float, maximum: float
    ) -> None:
        # Don't notify too often (unless it's the final notification)
        if value != maximum and time.time() - self._last_progress_reporting_time < 0.2:
            return

        self.send_message(
            BackendEvent(
                event_type="InlineProgress",
                command_id=cmd["id"],
                value=value,
                maximum=maximum,
                description=description,
            )
        )
        self._last_progress_reporting_time = time.time()

    def _report_current_action(self, cmd, description: str) -> None:
        self.send_message(
            BackendEvent(
                event_type="InlineProgress",
                command_id=cmd["id"],
                description=description,
            )
        )

    def _read_incoming_messages(self):
        # works in a separate thread
        while True:
            if not self._read_one_incoming_message():
                break

    def _read_one_incoming_message(self):
        msg_str = read_one_incoming_message_str(self._read_incoming_msg_line)
        if not msg_str:
            return False

        msg = parse_message(msg_str)
        if isinstance(msg, ImmediateCommand):
            # This will be handled right away
            self._handle_immediate_command(msg)
        else:
            self._incoming_message_queue.put(msg)
        return True

    def _prepare_command_response(
        self, response: Union[MessageFromBackend, Dict, None], command: CommandToBackend
    ) -> MessageFromBackend:
        if response is None:
            response = {}

        if "id" in command and "command_id" not in response:
            response["command_id"] = command["id"]

        if isinstance(response, MessageFromBackend):
            if "command_name" not in response:
                response["command_name"] = command["name"]
            return response
        else:
            if isinstance(response, dict):
                args = response
            else:
                args = {}

            if isinstance(command, ToplevelCommand):
                return ToplevelResponse(command_name=command.name, **args)
            else:
                assert isinstance(command, InlineCommand)
                return InlineResponse(command_name=command.name, **args)

    def send_message(self, msg: MessageFromBackend) -> None:
        sys.stdout.write(serialize_message(msg) + "\n")
        sys.stdout.flush()

    def _send_output(self, data, stream_name):
        if not data:
            return

        data = self._transform_output(data, stream_name)
        msg = BackendEvent(event_type="ProgramOutput", stream_name=stream_name, data=data)
        self._last_sent_output = data
        self.send_message(msg)

    def _transform_output(self, data, stream_name):
        return data

    def _read_incoming_msg_line(self) -> str:
        return sys.stdin.readline()

    def _perform_idle_tasks(self):
        """Executed when there is no commands in queue"""
        pass

    def _report_internal_exception(self, msg: str) -> None:
        user_msg = "PROBLEM IN turcar'S BACK-END: " + msg
        if sys.exc_info()[1]:
            err_msg = "\n".join(
                traceback.format_exception_only(sys.exc_info()[0], sys.exc_info()[1])
            ).strip()
            user_msg += f" ({err_msg})"

        user_msg += ".\nSee " + turcar.BACKEND_LOG_MARKER + " for more info."

        print(user_msg, file=sys.stderr)

    def _report_internal_warning(self, msg: str) -> None:
        user_msg = f"Warning: {msg}.\nSee backend.log for more info."
        print(user_msg, file=sys.stderr)

    @abstractmethod
    def _check_for_connection_error(self) -> None:
        ...

    @abstractmethod
    def _handle_user_input(self, msg: InputSubmission) -> None:
        pass

    @abstractmethod
    def _handle_eof_command(self, msg: EOFCommand) -> None:
        pass

    @abstractmethod
    def _handle_normal_command(self, cmd: CommandToBackend) -> None:
        pass

    @abstractmethod
    def _handle_immediate_command(self, cmd: ImmediateCommand) -> None:
        """Command handler will be executed in command reading thread, right after receiving the command"""


class MainBackend(BaseBackend, ABC):
    """Backend which does not forward to another backend"""

    def __init__(self):
        self._command_handlers = {}
        self._jedi_is_loaded = False
        BaseBackend.__init__(self)

    def add_command(self, command_name, handler):
        """Handler should be 1-argument function taking command object.

        Handler may return None (in this case no response is sent to frontend)
        or a BackendResponse
        """
        self._command_handlers[command_name] = handler

    def send_message(self, msg: MessageFromBackend) -> None:
        super().send_message(msg)

        # take the time for pre-loading jedi after the first toplevel response
        if isinstance(msg, ToplevelResponse):
            self._check_load_jedi()

    def _handle_normal_command(self, cmd: CommandToBackend) -> None:
        assert isinstance(cmd, (ToplevelCommand, InlineCommand))
        logger.debug("Command: %r", cmd)

        if cmd.name in self._command_handlers:
            handler = self._command_handlers[cmd.name]
        else:
            handler = getattr(self, "_cmd_" + cmd.name, None)

        if handler is None:
            if isinstance(cmd, ToplevelCommand):
                self._send_output(f"Unknown command '{cmd.name}'", "stderr")
            response = {"error": "Unknown command: " + cmd.name}
        else:
            try:
                response = handler(cmd)
                # Exceptions must be caused by turcar or plugins code, because the ones
                # from user code are caught at execution places
            except UserError as e:
                logger.info("UserError while handling %r", cmd.name, exc_info=True)
                if isinstance(cmd, ToplevelCommand):
                    print(str(e), file=sys.stderr)
                    response = {}
                else:
                    response = {"error": str(e)}
            except KeyboardInterrupt as e:
                if isinstance(cmd, ToplevelCommand):
                    print(str(e), file=sys.stderr)
                    response = {}
                else:
                    response = {"error": "Interrupted", "interrupted": True}
            except Exception as e:
                logger.exception("Exception while handling %r", cmd.name)
                self._report_internal_exception("Exception while handling %r" % cmd.name)
                sys.exit(1)

        if response is False:
            # Command doesn't want to send any response
            return

        real_response = self._prepare_command_response(response, cmd)
        self.send_message(real_response)

    def _cmd_get_dirs_children_info(self, cmd):
        """Provides information about immediate children of paths opened in a file browser"""
        data = {
            path: self._get_filtered_dir_children_info(path, cmd["include_hidden"])
            for path in cmd["paths"]
        }
        return {"node_id": cmd["node_id"], "dir_separator": self._get_sep(), "data": data}

    def _cmd_prepare_upload(self, cmd):
        """Returns info about items to be overwritten or merged by cmd.paths"""
        return {"existing_items": self._get_paths_info(cmd.target_paths, recurse=False)}

    def _cmd_prepare_download(self, cmd):
        assert "id" in cmd
        """Returns info about all items under and including cmd.paths"""
        return {"all_items": self._get_paths_info(cmd.source_paths, recurse=True)}

    def _cmd_shell_autocomplete(self, cmd):
        error = None
        try:
            from turcar import jedi_utils
        except ImportError:
            completions = []
            error = "Could not import jedi"
        else:
            import __main__

            with warnings.catch_warnings():
                completions = jedi_utils.get_interpreter_completions(
                    cmd.source, [__main__.__dict__], sys_path=self._get_sys_path_for_analysis()
                )

        return dict(
            source=cmd.source,
            completions=completions,
            error=error,
            row=cmd.row,
            column=cmd.column,
        )

    def _cmd_editor_autocomplete(self, cmd):
        logger.debug("Starting _cmd_editor_autocomplete")
        error = None
        try:
            from turcar import jedi_utils

            sys_path = self._get_sys_path_for_analysis()

            # add current dir for local files
            """
            if cmd.filename and is_local_path(cmd.filename):
                sys_path.insert(0, os.getcwd())
                logger.debug("editor autocomplete with %r", sys_path)
            """

            with warnings.catch_warnings():
                completions = jedi_utils.get_script_completions(
                    cmd.source,
                    cmd.row,
                    cmd.column,
                    cmd.filename,
                    sys_path=sys_path,
                )
        except ImportError:
            completions = []
            error = "Could not import jedi"

        return dict(
            source=cmd.source,
            row=cmd.row,
            column=cmd.column,
            filename=cmd.filename,
            completions=completions,
            error=error,
        )

    def _cmd_get_completion_details(self, cmd):
        # it is assumed this gets called after requesting editor or shell completions
        from turcar import jedi_utils

        return InlineResponse(
            "get_completion_details",
            full_name=cmd.full_name,
            details=jedi_utils.get_completion_details(cmd.full_name),
        )

    def _cmd_get_editor_calltip(self, cmd):
        from turcar import jedi_utils

        signatures = jedi_utils.get_script_signatures(
            cmd.source,
            cmd.row,
            cmd.column,
            cmd.filename,
            sys_path=self._get_sys_path_for_analysis(),
        )
        return InlineResponse(
            "get_editor_calltip",
            source=cmd.source,
            row=cmd.row,
            column=cmd.column,
            filename=cmd.filename,
            signatures=signatures,
        )

    def _cmd_get_shell_calltip(self, cmd):
        import __main__
        from turcar import jedi_utils

        signatures = jedi_utils.get_interpreter_signatures(
            cmd.source, [__main__.__dict__], sys_path=self._get_sys_path_for_analysis()
        )
        return InlineResponse(
            "get_shell_calltip",
            source=cmd.source,
            row=cmd.row,
            column=cmd.column,
            filename=cmd.filename,
            signatures=signatures,
        )

    def _cmd_highlight_occurrences(self, cmd):
        from turcar import jedi_utils

        refs = jedi_utils.get_references(
            cmd.source,
            cmd.row,
            cmd.column,
            cmd.filename,
            scope="file",
            sys_path=self._get_sys_path_for_analysis(),
        )

        return {"references": refs, "text_last_operation_time": cmd.text_last_operation_time}

    def _cmd_get_definitions(self, cmd):
        from turcar import jedi_utils

        defs = jedi_utils.get_definitions(
            cmd.source,
            cmd.row,
            cmd.column,
            filename=cmd.filename,
            sys_path=self._get_sys_path_for_analysis(),
        )
        return {"definitions": defs}

    def _cmd_get_active_distributions(self, cmd):
        raise NotImplementedError()

    def _cmd_install_distributions(self, cmd):
        raise NotImplementedError()

    def _cmd_uninstall_distributions(self, cmd):
        raise NotImplementedError()

    def _get_sys_path_for_analysis(self) -> Optional[List[str]]:
        return None

    def _get_paths_info(self, paths: List[str], recurse: bool) -> Dict[str, Dict]:
        result = {}

        for path in paths:
            info = self._get_path_info(path)
            if info is not None:
                info["anchor"] = path
                result[path] = info

            if recurse and info is not None and info["kind"] == "dir":
                desc_infos = self._get_dir_descendants_info(path)
                for key in desc_infos:
                    desc_infos[key]["anchor"] = path
                result.update(desc_infos)

        return result

    def _get_dir_descendants_info(self, path: str, include_hidden: bool = False) -> Dict[str, Dict]:
        """Assumes path is dir. Dict is keyed by full path"""
        result = {}
        children_info = self._get_filtered_dir_children_info(path, include_hidden)
        for child_name, child_info in children_info.items():
            full_child_path = path + self._get_sep() + child_name
            result[full_child_path] = child_info
            if child_info["kind"] == "dir":
                result.update(self._get_dir_descendants_info(full_child_path))

        return result

    def _get_filtered_dir_children_info(
        self, path: str, include_hidden: bool = False
    ) -> Optional[Dict[str, Dict]]:
        children = self._get_dir_children_info(path, include_hidden)
        if children is None:
            return None

        return {name: children[name] for name in children if name not in IGNORED_FILES_AND_DIRS}

    @abstractmethod
    def _get_path_info(self, path: str) -> Optional[Dict]:
        """Returns information about this path or None if it doesn't exist"""

    @abstractmethod
    def _get_dir_children_info(
        self, path: str, include_hidden: bool = False
    ) -> Optional[Dict[str, Dict]]:
        """For existing dirs returns Dict[child_short_name, Dict of its information].
        Returns None if path doesn't exist or is not a dir.
        """

    @abstractmethod
    def _get_sep(self) -> str:
        """Returns symbol for combining parent directory path and child name"""

    def _check_load_jedi(self) -> None:
        if self._jedi_is_loaded:
            return
        logger.info("Loading Jedi")

        report_time("Before loading Jedi")
        try_load_modules_with_frontend_sys_path(["jedi", "parso"])
        self._jedi_is_loaded = True
        report_time("After loading Jedi")


class UploadDownloadMixin(ABC):
    """Backend, which runs on a local process and talks to a nonlocal system,
    and therefore is able to upload/download"""

    def _cmd_download(self, cmd):
        errors = self._transfer_files_and_dirs(
            cmd.items, self._ensure_local_directory, self._download_file, cmd, pathlib.Path
        )
        return {"errors": errors}

    def _cmd_upload(self, cmd):
        def upload_file_wrapper(source_path, target_path, callback):
            self._upload_file(
                source_path, target_path, callback, cmd["make_shebang_scripts_executable"]
            )

        errors = self._transfer_files_and_dirs(
            cmd.items,
            self._ensure_remote_directory,
            upload_file_wrapper,
            cmd,
            pathlib.PurePosixPath,
        )
        return {"errors": errors}

    def _cmd_read_file(self, cmd):
        def callback(completed, total):
            self._report_progress(cmd, cmd["path"], completed, total)

        with io.BytesIO() as fp:
            self._read_file(cmd["path"], fp, callback)
            fp.seek(0)
            content_bytes = fp.read()

        return {"content_bytes": content_bytes, "path": cmd["path"]}

    def _cmd_write_file(self, cmd):
        def callback(completed, total):
            self._report_progress(cmd, cmd["path"], completed, total)

        with io.BytesIO() as fp:
            fp.write(cmd["content_bytes"])
            fp.seek(0)
            self._write_file(
                fp,
                cmd["path"],
                file_size=len(cmd["content_bytes"]),
                callback=callback,
                make_shebang_scripts_executable=cmd["make_shebang_scripts_executable"],
            )

        return InlineResponse(
            command_name="write_file", path=cmd["path"], editor_id=cmd.get("editor_id")
        )

    def _supports_directories(self) -> bool:
        return True

    def _transfer_files_and_dirs(
        self,
        items: Iterable[Dict],
        ensure_dir_fun: Callable[[str], None],
        transfer_file_fun: Callable,
        cmd,
        target_path_class,
    ) -> List[str]:
        total_cost = 0
        for item in items:
            if item["kind"] == "file":
                total_cost += item["size"] + self._get_file_fixed_cost()
            else:
                total_cost += self._get_dir_transfer_cost()

        completed_cost = 0
        errors = []

        ensured_dirs = set()

        def ensure_dir(path):
            if path in ensured_dirs:
                return
            ensure_dir_fun(path)
            ensured_dirs.add(path)

        for item in sorted(items, key=lambda x: x["source_path"]):
            self._report_progress(cmd, "Starting", completed_cost, total_cost)

            def copy_bytes_notifier(completed_bytes, total_bytes):
                completed = completed_cost + completed_bytes
                desc = str(round(completed / total_cost * 100)) + "%"

                self._report_progress(cmd, desc, completed, total_cost)

            try:
                if item["kind"] == "dir":
                    ensure_dir(item["target_path"])
                    completed_cost += self._get_dir_transfer_cost()
                else:
                    if self._supports_directories():
                        ensure_dir(self._get_parent_directory(item["target_path"]))
                    print("%s (%d bytes)" % (item["source_path"], item["size"]))
                    transfer_file_fun(item["source_path"], item["target_path"], copy_bytes_notifier)
                    completed_cost += self._get_file_fixed_cost() + item["size"]
            except OSError as e:
                errors.append(
                    "Could not copy %s to %s: %s"
                    % (item["source_path"], item["target_path"], str(e))
                )

        return errors

    def _download_file(self, source_path, target_path, callback):
        with open(target_path, "bw") as target_fp:
            self._read_file(source_path, target_fp, callback)

    def _upload_file(
        self, source_path, target_path, callback, make_shebang_scripts_executable: bool
    ):
        with open(source_path, "br") as source_fp:
            self._write_file(
                source_fp,
                target_path,
                os.path.getsize(source_path),
                callback,
                make_shebang_scripts_executable,
            )

    def _get_dir_transfer_cost(self):
        # Validating and maybe creating a directory is taken to be equal to copying this number of bytes
        return 1000

    def _get_file_fixed_cost(self):
        # Creating or overwriting a file is taken to be equal to copying this number of bytes
        return 100

    def _get_parent_directory(self, path: str):
        return universal_dirname(path)

    def _ensure_local_directory(self, path: str) -> None:
        os.makedirs(path, NEW_DIR_MODE, exist_ok=True)

    def _ensure_remote_directory(self, path: str) -> None:
        # assuming remote system is Posix
        ensure_posix_directory(path, self._get_stat_mode_for_upload, self._mkdir_for_upload)

    @abstractmethod
    def _get_stat_mode_for_upload(self, path: str) -> Optional[int]:
        """returns None if path doesn't exist"""

    @abstractmethod
    def _mkdir_for_upload(self, path: str) -> None:
        raise NotImplementedError()

    @abstractmethod
    def _write_file(
        self,
        source_fp: BinaryIO,
        target_path: str,
        file_size: int,
        callback: Callable[[int, int], None],
        make_shebang_scripts_executable: bool,
    ) -> None:
        raise NotImplementedError()

    @abstractmethod
    def _read_file(
        self, source_path: str, target_fp: BinaryIO, callback: Callable[[int, int], None]
    ) -> None:
        raise NotImplementedError()

    @abstractmethod
    def _report_internal_exception(self):
        raise NotImplementedError()

    @abstractmethod
    def _report_progress(
        self, cmd, description: Optional[str], value: float, maximum: float
    ) -> None:
        raise NotImplementedError()


class RemoteProcess:
    """Modelled after subprocess.Popen"""

    def __init__(self, client, channel, stdin, stdout, pid):
        self._client = client
        self._channel = channel
        self.stdin = stdin
        self.stdout = stdout
        self.pid = pid
        self.returncode = None

    def poll(self):
        if self._channel.exit_status_ready():
            self.returncode = self._channel.recv_exit_status()
            return self.returncode
        else:
            return None

    def wait(self):
        self.returncode = self._channel.recv_exit_status()
        return self.returncode

    def kill(self):
        _, stdout, _ = self._client.exec_command("kill -9 %s" % self.pid)
        # wait until completion
        stdout.channel.recv_exit_status()


class SshMixin(UploadDownloadMixin):
    def __init__(self, host, user, password, interpreter, cwd):
        # UploadDownloadMixin.__init__(self)
        execute_with_frontend_sys_path(self._try_load_paramiko)
        import paramiko
        from paramiko.client import AutoAddPolicy, SSHClient

        self._host = host
        self._user = user
        self._password = password
        self._target_interpreter = interpreter
        self._cwd = cwd
        self._proc = None  # type: Optional[RemoteProcess]
        self._sftp = None  # type: Optional[paramiko.SFTPClient]
        self._client = SSHClient()
        self._client.load_system_host_keys()
        self._client.set_missing_host_key_policy(paramiko.client.AutoAddPolicy())
        # TODO: does it get closed properly after process gets killed?
        self._connect()

    def _try_load_paramiko(self):
        try:
            import paramiko.client
        except ImportError:
            logger.info("Could not import paramiko")
            print(
                "\nThis back-end requires an extra package named 'paramiko'."
                " Install it from 'Tools => Manage plug-ins' or via your system package manager.",
                file=sys.stderr,
            )
            sys.exit(1)

    def _connect(self):
        from paramiko import SSHException

        try:
            self._client.connect(
                hostname=self._host,
                username=self._user,
                password=self._password,
                passphrase=self._password,
            )
        except (SSHException, OSError) as e:
            print(
                "\nCan't connect to '%s' with user '%s': %s" % (self._host, self._user, str(e)),
                file=sys.stderr,
            )
            print("Re-check your host, authentication method, password or keys.", file=sys.stderr)
            delete_stored_ssh_password()

            sys.exit(1)

    def _create_remote_process(self, cmd_items: List[str], cwd: str, env: Dict) -> RemoteProcess:
        import shlex

        # Before running the main thing:
        # * print process id (so that we can kill it later)
        #   http://redes-privadas-virtuales.blogspot.com/2013/03/getting-hold-of-remote-pid-through.html
        # * change to desired directory
        #
        # About -onlcr: https://stackoverflow.com/q/35887380/261181
        cmd_line_str = (
            "echo $$ ; stty -echo ; stty -onlcr ; "
            + (" cd %s  2> /dev/null ;" % shlex.quote(cwd) if cwd else "")
            + (" exec " + " ".join(map(shlex.quote, cmd_items)))
        )
        stdin, stdout, _ = self._client.exec_command(
            cmd_line_str, bufsize=0, get_pty=True, environment=env
        )

        # stderr gets directed to stdout because of pty
        pid = stdout.readline().strip()
        ack = stdout.readline().strip()
        if ack != PROCESS_ACK:
            raise RuntimeError(f"Got {ack!r} instead of expected {PROCESS_ACK!r}")
        channel = stdout.channel

        return RemoteProcess(self._client, channel, stdin, stdout, pid)

    def _handle_immediate_command(self, cmd: ImmediateCommand) -> None:
        if cmd.name == "kill":
            self._kill()
        elif cmd.name == "interrupt":
            self._interrupt()
        else:
            raise RuntimeError("Unknown immediateCommand %s" % cmd.name)

    def _kill(self):
        if self._proc is None or self._proc.poll() is not None:
            return

        self._proc.kill()

    def _interrupt(self):
        pass

    def _get_sftp(self, fresh: bool):
        if fresh and self._sftp is not None:
            self._sftp.close()
            self._sftp = None

        if self._sftp is None:
            import paramiko

            # TODO: does it get closed properly after process gets killed?
            self._sftp = paramiko.SFTPClient.from_transport(self._client.get_transport())

        return self._sftp

    def _read_file(
        self, source_path: str, target_fp: BinaryIO, callback: Callable[[int, int], None]
    ) -> None:
        self._perform_sftp_operation_with_retry(
            lambda sftp: sftp.getfo(source_path, target_fp, callback)
        )

    def _write_file(
        self,
        source_fp: BinaryIO,
        target_path: str,
        file_size: int,
        callback: Callable[[int, int], None],
        make_shebang_scripts_executable: bool,
    ) -> None:
        logger.info("Writing bytes to %r", target_path)
        if make_shebang_scripts_executable:
            source_fp, has_shebang = convert_newlines_if_has_shebang(source_fp)
        else:
            has_shebang = None

        self._perform_sftp_operation_with_retry(
            lambda sftp: sftp.putfo(source_fp, target_path, callback)
        )

        logger.debug(
            "make_shebang_scripts_executable: %r, has_shebang: %r",
            make_shebang_scripts_executable,
            has_shebang,
        )
        if make_shebang_scripts_executable and has_shebang:
            self._perform_sftp_operation_with_retry(lambda sftp: sftp.chmod(target_path, 0o755))

    def _perform_sftp_operation_with_retry(self, operation) -> Any:
        try:
            return operation(self._get_sftp(fresh=False))
        except OSError:
            # It looks like SFTPClient gets stale after a while.
            # Try again with fresh SFTPClient
            return operation(self._get_sftp(fresh=True))

    def _get_stat_mode_for_upload(self, path: str) -> Optional[int]:
        try:
            return self._perform_sftp_operation_with_retry(lambda sftp: sftp.stat(path).st_mode)
        except OSError as e:
            return None

    def _mkdir_for_upload(self, path: str) -> None:
        self._perform_sftp_operation_with_retry(lambda sftp: sftp.mkdir(path, NEW_DIR_MODE))


def _longest_common_path_prefix(str_paths, path_class):
    assert str_paths

    if len(str_paths) == 1:
        return str_paths[0]

    list_of_parts = []
    for str_path in str_paths:
        list_of_parts.append(path_class(str_path).parts)

    first = list_of_parts[0]
    rest = list_of_parts[1:]

    i = 0
    while i < len(first):
        item_i = first[i]
        if not all([len(x) > i and x[i] == item_i for x in rest]):
            break
        else:
            i += 1

    if i == 0:
        return ""

    result = path_class(first[0])
    for j in range(1, i):
        result = result.joinpath(first[j])

    return str(result)


def ensure_posix_directory(
    path: str, stat_mode_fun: Callable[[str], Optional[int]], mkdir_fun: Callable[[str], None]
) -> None:
    assert path.startswith("/")
    if path == "/":
        return

    for step in list(reversed(list(map(str, pathlib.PurePosixPath(path).parents)))) + [path]:
        if step != "/":
            mode = stat_mode_fun(step)
            if mode is None:
                mkdir_fun(step)
            elif not stat.S_ISDIR(mode):
                raise AssertionError("'%s' is file, not a directory" % step)


def interrupt_local_process() -> None:
    """Meant to be executed from a background thread"""
    import signal

    if hasattr(signal, "raise_signal"):
        # Python 3.8 and later
        signal.raise_signal(signal.SIGINT)
    elif sys.platform == "win32":
        # https://stackoverflow.com/a/51122690/261181
        import ctypes

        ucrtbase = ctypes.CDLL("ucrtbase")
        c_raise = ucrtbase["raise"]
        c_raise(signal.SIGINT)
    else:
        # Does not give KeyboardInterrupt in Windows
        os.kill(os.getpid(), signal.SIGINT)


def get_ssh_password_file_path():
    from turcar import turcar_USER_DIR

    return os.path.join(turcar_USER_DIR, "ssh_password")


def delete_stored_ssh_password():
    if os.path.exists(get_ssh_password_file_path()):
        # invalidate stored password
        os.remove(get_ssh_password_file_path())


def convert_newlines_if_has_shebang(fp: BinaryIO) -> Tuple[BinaryIO, bool]:
    if fp.read(3) == b"#!/":
        fp.seek(0)
        new_fp = io.BytesIO()
        new_fp.write(fp.read().replace(b"\r\n", b"\n"))
        fp.close()
        new_fp.seek(0)
        return new_fp, True
    else:
        fp.seek(0)
        return fp, False


if __name__ == "__main__":
    # print(_closest_common_directory(["C:\\kala\\pala", "C:\\kala", "D:\\kuku"], pathlib.PureWindowsPath))
    print(_longest_common_path_prefix(["C:\\kala\\pala", "C:\\kala"], pathlib.PureWindowsPath))
    print(_longest_common_path_prefix(["C:\\kala\\pala"], pathlib.PureWindowsPath))



import re
import tkinter as tk
from tkinter import ttk

from turcar import get_workbench
from turcar.languages import tr
from turcar.ui_utils import SafeScrollbar


class OutlineView(ttk.Frame):
    def __init__(self, master):
        ttk.Frame.__init__(self, master)
        self._init_widgets()

        self._tab_changed_binding = (
            get_workbench()
            .get_editor_notebook()
            .bind("<<NotebookTabChanged>>", self._update_frame_contents, True)
        )
        get_workbench().bind("Save", self._update_frame_contents, True)
        get_workbench().bind("SaveAs", self._update_frame_contents, True)
        get_workbench().bind_class("Text", "<<NewLine>>", self._update_frame_contents, True)

        self._update_frame_contents()

    def destroy(self):
        try:
            # Not sure if editor notebook is still living
            get_workbench().get_editor_notebook().unbind(
                "<<NotebookTabChanged>>", self._tab_changed_binding
            )
        except Exception:
            pass
        self.vert_scrollbar["command"] = None
        ttk.Frame.destroy(self)

    def _init_widgets(self):
        # init and place scrollbar
        self.vert_scrollbar = SafeScrollbar(self, orient=tk.VERTICAL)
        self.vert_scrollbar.grid(row=0, column=1, sticky=tk.NSEW)

        # init and place tree
        self.tree = ttk.Treeview(self, yscrollcommand=self.vert_scrollbar.set)
        self.tree.grid(row=0, column=0, sticky=tk.NSEW)
        self.vert_scrollbar["command"] = self.tree.yview

        # set single-cell frame
        self.columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)

        # init tree events
        self.tree.bind("<<TreeviewSelect>>", self._on_select, True)
        self.tree.bind("<Map>", self._update_frame_contents, True)

        # configure the only tree column
        self.tree.column("#0", anchor=tk.W, stretch=True)
        # self.tree.heading('#0', text='Item (type @ line)', anchor=tk.W)
        self.tree["show"] = ("tree",)

        self._class_img = get_workbench().get_image("outline-class")
        self._method_img = get_workbench().get_image("outline-method")

    def _update_frame_contents(self, event=None):
        if not self.winfo_ismapped():
            return

        self._clear_tree()

        editor = get_workbench().get_editor_notebook().get_current_editor()
        if editor is None:
            return

        root = self._parse_source(editor.get_code_view().get_content())
        for child in root[2]:
            self._add_item_to_tree("", child)

    def _parse_source(self, source):
        # all nodes in format (parent, node_indent, node_children, name, type, linenumber)
        root_node = (None, 0, [], None, None, None)  # name, type and linenumber not needed for root
        active_node = root_node

        lineno = 0
        for line in source.split("\n"):
            lineno += 1
            m = re.match(r"[ ]*[\w]{1}", line)
            if m:
                indent = len(m.group(0))
                while indent <= active_node[1]:
                    active_node = active_node[0]

                t = re.match(
                    r"[ \t]*(async[ \t]+)?(?P<type>(def|class){1})[ ]+(?P<name>[\w]+)", line
                )
                if t:
                    current = (active_node, indent, [], t.group("name"), t.group("type"), lineno)
                    active_node[2].append(current)
                    active_node = current

        return root_node

    # adds a single item to the tree, recursively calls itself to add any child nodes
    def _add_item_to_tree(self, parent, item):
        # create the text to be played for this item
        item_type = item[4]
        item_text = " " + item[3]

        if item_type == "class":
            image = self._class_img
        elif item_type == "def":
            image = self._method_img
        else:
            image = None

        # insert the item, set lineno as a 'hidden' value
        current = self.tree.insert(parent, "end", text=item_text, values=item[5], image=image)

        for child in item[2]:
            self._add_item_to_tree(current, child)

    # clears the tree by deleting all items
    def _clear_tree(self):
        for child_id in self.tree.get_children():
            self.tree.delete(child_id)

    def _on_select(self, event):
        editor = get_workbench().get_editor_notebook().get_current_editor()
        if editor:
            code_view = editor.get_code_view()
            focus = self.tree.focus()
            if not focus:
                return

            values = self.tree.item(focus)["values"]
            if not values:
                return

            lineno = values[0]
            index = code_view.text.index(str(lineno) + ".0")
            code_view.text.see(index)  # make sure that the double-clicked item is visible
            code_view.text.select_lines(lineno, lineno)

            get_workbench().event_generate(
                "OutlineDoubleClick", item_text=self.tree.item(self.tree.focus(), option="text")
            )


def load_plugin() -> None:
    get_workbench().add_view(OutlineView, tr("Outline"), "ne")



import urllib.parse
from logging import getLogger
from tkinter import messagebox
from typing import Any, Dict, List, Optional

from turcar.languages import tr
from turcar.misc_utils import levenshtein_distance
from turcar.plugins.micropython import LocalMicroPythonProxy, MicroPythonProxy
from turcar.plugins.pip_gui import BackendPipDialog, get_not_supported_translation

MICROPYTHON_ORG_JSON = "https://micropython.org/pi/v2/index.json"

logger = getLogger(__name__)


class MicroPythonPipDialog(BackendPipDialog):
    def __init__(self, master):
        self._mp_org_index_data = None
        super().__init__(master)
        assert isinstance(self._backend_proxy, MicroPythonProxy)

    def _confirm_install(self, package_data: Dict) -> bool:
        if not self._looks_like_micropython_package(package_data):
            if not messagebox.askyesno(
                title=tr("Confirmation"),
                message=tr(
                    "This doesn't look like MicroPython/CircuitPython package.\n"
                    "Are you sure you want to install it?"
                ),
                parent=self,
            ):
                return False

        return super()._confirm_install(package_data)

    def _looks_like_micropython_package(self, package_data: Dict) -> bool:
        if package_data.get("mp_org", False):
            return True

        name = package_data["info"]["name"]
        for token in ["micropython", "circuitpython", "pycopy"]:
            if token in name.lower():
                return True

        classifiers = package_data["info"].get("classifiers", [])
        logger.debug("package classifiers: %s", classifiers)
        for mp_class in [
            "Programming Language :: Python :: Implementation :: MicroPython",
            "Programming Language :: Python :: Implementation :: CircuitPython",
        ]:
            if mp_class in classifiers:
                return True

        return False

    def _get_package_metadata_url(self, name: str, version_str: Optional[str]) -> str:
        return "https://micropython.org/pi/{}/json".format(urllib.parse.quote(name))

    def _get_package_metadata_fallback_url(
        self, name: str, version_str: Optional[str]
    ) -> Optional[str]:
        return super()._get_package_metadata_url(name, version_str)

    def get_search_button_text(self):
        return tr("Search micropython-lib and PyPI")

    def _download_package_info(self, name: str, version_str: Optional[str]) -> Dict:
        # Try mp.org first
        index_data = self._get_mp_org_index_data()

        for package in index_data["packages"]:
            if self._normalize_name(package["name"]) == self._normalize_name(name):
                info = {
                    "name": package["name"],
                    "version": package["version"],
                }
                if package.get("author"):
                    info["author"] = package["author"]
                if package.get("license"):
                    info["license"] = package["license"]
                if package.get("description"):
                    info["summary"] = package["description"]
                # TODO: deps?

                releases = {ver: [] for ver in package["versions"]["py"]}

                return {"info": info, "releases": releases, "mp_org": True}

        return super()._download_package_info(name, version_str)

    def _get_mp_org_index_data(self) -> Dict[str, Any]:
        if not self._mp_org_index_data:
            import json
            from urllib.request import urlopen

            with urlopen(MICROPYTHON_ORG_JSON, timeout=10) as fp:
                self._mp_org_index_data = json.load(fp)

        return self._mp_org_index_data

    def _get_target_directory(self):
        # TODO: should this be pipkin's decision?
        target_dir = self._backend_proxy.get_pip_target_dir()
        return target_dir

    def _is_read_only(self):
        return self._get_target_directory() is None

    def does_support_update_deps_switch(self):
        return False

    def _use_user_install(self):
        return False

    def _installer_runs_locally(self):
        return True

    def _show_instructions_about_target(self):
        self._append_info_text(tr("Target") + "\n", ("caption",))

        if isinstance(self._backend_proxy, LocalMicroPythonProxy):
            dir_tags = ("url",)
        else:
            dir_tags = ()

        self._append_info_text(self._get_target_directory(), dir_tags)
        self._append_info_text("\n")

    def _show_extra_instructions(self):
        self._append_info_text("\n\n")
        self._append_info_text("Under the hood " + "\n", ("caption", "right"))
        self._append_info_text(
            "This dialog uses `pipkin`, a new command line tool for managing "
            "MicroPython and CircuitPython packages. ",
            ("right",),
        )
        self._append_info_text("See ", ("right",))
        self._append_info_text("https://pypi.org/project/pipkin/", ("url", "right"))
        self._append_info_text(" for more info. \n", ("right",))

    def _show_read_only_instructions(self):
        self._append_info_text(tr("Not available") + "\n", ("caption",))
        if not self._get_target_directory():
            reason = " (" + tr("no absolute lib directory in sys.path") + ")"
        else:
            reason = ""
        self.info_text.direct_insert(
            "end",
            get_not_supported_translation() + reason + "\n\n",
        )

    def _fetch_search_results(self, query: str) -> Dict[str, List[Dict[str, str]]]:
        """Will be executed in background thread"""

        mp_org_result = self._perform_micropython_org_search(query)
        pypi_result = self._perform_pypi_search(query, source="PyPI")

        combined_result = []
        mp_org_names = set()

        for item in mp_org_result:
            combined_result.append(item)
            mp_org_names.add(item["name"])

        for item in pypi_result:
            # TODO: normalize?
            if item["name"] not in mp_org_names:
                # add slight penalty for non-perfect PyPI matches, as most users look for mp.org results
                if item["distance"] != 0:
                    item["distance"] += 1
                combined_result.append(item)

        sorted_result = sorted(combined_result, key=lambda x: x["distance"])
        filtered_result = filter(lambda x: x["distance"] < 5, sorted_result[:20])

        # Combining, because this makes the shadowing performed by Pipkin more clear
        return {"combined": list(filtered_result)}

    def _perform_micropython_org_search(self, query: str) -> List[Dict[str, str]]:
        logger.info("Searching %r for %r", MICROPYTHON_ORG_JSON, query)
        data = self._get_mp_org_index_data()

        result = []
        for package in data["packages"]:
            result.append(
                {
                    "name": package["name"],
                    "description": package["description"] or None,
                    "source": "micropython-lib",
                    "distance": levenshtein_distance(query, package["name"]),
                }
            )

        logger.info("Got %r items", len(result))
        return result

    def _get_interpreter_description(self):
        return self._backend_proxy.get_full_label()



from pybricks.__stub.__control import Control
from pybricks.parameters import Direction, Port, Stop
from typing import List, Union


class Motor:
    """
    Generic class to control motors with built-in rotation sensors.

    Attributes:
        control (Control): The motor's PID control. 

    Args:
        port (Port): Port to which the motor is connected.
        positive_direction (Direction): Which direction the motor should turn when you give a positive speed or angle.
        gears (Union[List[int], List[List[int]]]): List of gears linked to the motor. For example, [12, 36] represents a gear train with a 12-tooth and a 36-tooth gear. Use a list of lists for multiple gear trains, such as [[12, 36], [20, 16, 40]].

    Note:
        When you specify a gear train, all motor commands and settings are automatically adjusted to account for the resulting gear ratio. The motor direction remains unchanged by this.
    """

    def __init__(self, port: Port, positive_direction: Direction = Direction.CLOCKWISE, gears: Union[List[int], List[List[int]]] = None):
        if port == Port.S1 or port == Port.S2 or port == port.S3 or port == port.S4:
            raise ValueError("Motors must use Port A, B, C, or D.")
        self.control = Control()  # type: Control

    def speed(self) -> int:
        """
        Gets the speed of the motor.

        Returns:
            Motor speed in degrees/second.
        """
        return 0

    def angle(self) -> int:
        """
        Get the rotation angle of the motor.

        Returns:
            Motor angle in degrees.
        """
        return 0

    def reset_angle(self, angle: int):
        """
        Sets the accumulated rotation angle of the motor to a desired value.

        Args:
            angle (int): Value to which the angle should be reset in degrees.
        """
        ...

    def stop(self):
        """
        Stops the motor and lets it spin freely.

        The motor gradually stops due to friction.
        """
        ...

    def brake(self):
        """
        Passively brakes the motor.

        The motor stops due to friction, plus the voltage that is generated while the motor is still moving.
        """
        ...

    def hold(self):
        """
        Stops the motor and actively holds it at its current angle.
        """
        ...

    def run(self, speed: int):
        """
        Runs the motor at a constant speed.

        The motor accelerates to the given speed and keeps running at this speed until you give a new command.

        Args:
            speed (int): Speed of the motor in degrees/second.
        """
        ...

    def run_time(self, speed: int, time: int, then: Stop = Stop.HOLD, wait: bool = True):
        """
        Runs the motor at a constant speed for a given amount of time.

        The motor accelerates to the given speed, keeps running at this speed, and then decelerates. The total maneuver lasts for exactly the given amount of time.

        Args:
            speed (int): Speed of the motor in degrees/second.
            time (int): Duration of the maneuver in milliseconds.
            then (Stop): What to do after coming to a standstill.
            wait (bool): Wait for the maneuver to complete before continuing with the rest of the program.
        """
        ...

    def run_angle(self, speed: int, rotation_angle: int, then: Stop = Stop.HOLD, wait: bool = True):
        """
        Run the motor at a constant speed by a given angle. 

        Args:
            speed (int): Speed of the motor in degrees/second.
            rotation_angle (int): Angle by which the motor should rotate in degrees.
            then (Stop): WWhat to do after coming to a standstill.
            wait (bool): Wait for the motor to reach the target before continuing with the rest of the program.
        """
        ...

    def run_target(self, speed: int, target_angle: int, then: Stop = Stop.COAST, wait: bool = True):
        """
        Runs the motor at a constant speed towards a given target angle.

        The direction of rotation is automatically selected based on the target angle. It does matter if speed is positive or negative.

        Args:
            speed (int): Speed of the motor in degrees/second.
            target_angle (int): Angle that the motor should rotate to in degrees.
            then (Stop): What to do after coming to a standstill.
            wait (bool): Wait for the motor to reach the target before continuing with the rest of the program.
        """
        ...

    def run_until_stalled(self, speed: int, then: Stop = Stop.COAST, duty_limit: int = None) -> int:
        """
        Runs the motor at a constant speed until it stalls.

        Args:
            speed (int): Speed of the motor in degrees/second.
            then (Stop): What to do after coming to a standstill.
            duty_limit (int) – Torque limit during this command as a percentage (0 - 100). This is useful to avoid applying the full motor torque to a geared or lever mechanism.

        Returns:	
            Angle at which the motor becomes stalled in degrees.
        """
        return 0

    def dc(self, duty: int):
        """
        Rotates the motor at a given duty cycle (also known as “power”).

        This method lets you use a motor just like a simple DC motor.

        Args:
            duty (int): The duty cycle as a percentage (-100 to 100).
        """
        ...

    def track_target(self, target_angle: int):
        """
        Tracks a target angle. This is similar to run_target(), but the usual smooth acceleration is skipped: it will move to the target angle as fast as possible. This method is useful if you want to continuously change the target angle.

        Args:
            target_angle (int): Target angle that the motor should rotate to in degrees.
        """
        ...



# def fibonacci(n):
#     # if n <= 0:
#     #     return []
#     # elif n == 1:
#     #     return [0]
#     # elif n == 2:
#     #     return [0, 1]
#     if n <= 2:
#         return [1, 1]
#     else:
#         fib_sequence = fibonacci(n - 1)
#         fib_sequence.append(fib_sequence[-1] + fib_sequence[-2])
#         return fib_sequence
#
# # 测试
# n = 10
# sequence = fibonacci(n)
# print(sequence)

def fibonacci(n):
    if n <= 0:
        return []
    elif n == 1:
        return [1]
    elif n == 2:
        return [1, 1]
    else:
        fib_sequence = [1, 1] + [0] * (n - 2)  # 初始化斐波那契数列数组
        for i in range(2, n):
            fib_sequence[i] = fib_sequence[i - 1] + fib_sequence[i - 2]
        return fib_sequence

# 测试
n = 10
sequence = fibonacci(n)
print(sequence)




from pybricks.media.ev3dev import Font, ImageFile
from pybricks.parameters import Color
from typing import Union

empty = None
class Image:
    """
    Object representing a graphics image. This can either be an in-memory copy of an image or the image displayed on a screen.

    Args:
        source (str, Image, or ImageFile): The source of the image.
        sub (bool): If sub is True, then the image object will act as a sub-image of the source image (this only works if the type of source is Image and not when it is a str or ImageFile).

    Attributes:
        width: The width of this image in pixels.
        height: The height of this image in pixels.

    Note:
        If source is a string or ImageFile, then the image will be loaded from the file path given by the string. Only .png files are supported. As a special case, if the string is '_screen_', the image will be configured to draw directly on the screen.

        If an Image is given, the new object will contain a copy of the source image object.

        Additional keyword arguments x1, y1, x2, y2 are needed when sub=True. These specify the top-left and bottom-right coordinates in the source image that will be used as the bounds for the sub-image.
    """

    def __init__(self, source: Union['Image', str, ImageFile], *, sub: bool = False, x1: int = None, y1: int = None, x2: int = None, y2: int = None):
        self.width = 0  # type: int
        self.height = 0  # type: int

    @staticmethod
    def empty(width: int = 178, height: int = 128) -> 'Image':
        """
        Creates a new empty Image object.

        Args:
            width (int): The width of the image in pixels.
            height (int): The height of the image in pixels.

        Returns:
            A new image with all pixels set to Color.WHITE.

        Raises:
            TypeError: width or height is not a number
            ValueError: width or height is less than 1.
            RuntimeError: There was a problem allocating a new image.
        """
        return empty

    def draw_text(self, x: int, y: int, text: str, text_color: Color = Color.BLACK, background_color: Color = None):
        """
        Draws text on this image.

        The most recent font set using set_font() will be used or Font.DEFAULT if no font has been set yet.

        Args:
            x (int): The x-axis value where the left side of the text will start.
            y (int): The y-axis value where the top of the text will start.
            text (str): The text to draw.
            text_color (Color): The color used for drawing the text.
            background_color (Color): The color used to fill the rectangle behind the text or None for transparent background.
        """
        ...

    def print(self, *args, sep: str = "", end: str = "\n"):
        """
        Prints a line of text on this image.

        This method works like the builtin print() function, but it writes on this image instead.

        You can set the font using set_font(). If no font has been set, Font.DEFAULT will be used. The text is always printed used black text with a white background.

        Unlike the builtin print(), the text does not wrap if it is too wide to fit on this image. It just gets cut off. But if the text would go off of the bottom of this image, the entire image is scrolled up and the text is printed in the new blank area at the bottom of this image.

        Args:
            args (object): Zero or more objects to print.
            sep (str): Separator that will be placed between each object that is printed.
            end (str): End of line that will be printed after the last object.
        """
        ...

    def set_font(self, font: Font):
        """
        Sets the font used for writing on this image.

        The font is used for both draw_text() and print().

        Args:
            font (Font): The font to use.
        """
        ...

    def draw_image(self, x: int, y: int, source: Union[str, 'Image', ImageFile], transparent: Color = None):
        """
        Draws the source image on this image.

        Args:
            x (int): The x-axis value where the left side of the image will start.
            y (int): The y-axis value where the top of the image will start.
            source (Image, ImageFile, or str): The source Image. If the argument is a string, then the source image is loaded from file.
            transparent (Color): The color of image to treat as transparent or None for no transparency.
        """
        ...

    def draw_pixel(self, x: int, y: int, color: Color = Color.BLACK):
        """
        Draws a single pixel on this image.

        Args:
            x (int): The x coordinate of the pixel.
            y (int): The y coordinate of the pixel.
            color (Color): The color of the pixel.
        """
        ...

    def draw_line(self, x1: int, y1: int, x2: int, y2: int, width: int = 1, color: Color = Color.BLACK):
        """
        Draws a line on this image.

        Args:
            x1 (int): The x coordinate of the starting point of the line.
            y1 (int): The y coordinate of the starting point of the line.
            x2 (int): The x coordinate of the ending point of the line.
            y2 (int): The y coordinate of the ending point of the line.
            width (int): The width of the line in pixels.
            color (Color): The color of the line.
        """
        ...

    def draw_box(self, x1: int, y1: int, x2: int, y2: int, r: int = 0, fill: bool = False, color: Color = Color.BLACK):
        """
        Draws a box on this image.

        Args:
            x1 (int): The x coordinate of the left side of the box.
            y1 (int): The y coordinate of the top of the box.
            x2 (int): The x coordinate of the right side of the box.
            y2 (int): The y coordinate of the bottom of the box.
            r (int): The radius of the corners of the box.
            fill (bool): If True, the box will be filled with color, otherwise only the outline of the box will be drawn.
            color (Color): The color of the box.
        """
        ...

    def draw_circle(self, x: int, y: int, r: int, fill: bool = False, color: Color = Color.BLACK):
        """
        Draws a circle on this image.

        Args:
            x (int): The x coordinate of the center of the circle.
            y (int): The y coordinate of the center of the circle.
            r (int): The radius of the circle.
            fill (bool): If True, the circle will be filled with color, otherwise only the circumference will be drawn.
            color (Color): The color of the circle.
        """
        ...

    def clear(self):
        """
        Clears this image. All pixels on this image will be set to Color.WHITE.
        """
        ...

    def load_image(self, source: Union['Image', str, ImageFile]):
        """
        Clears this image, then draws the source image centered in this image.

        Args:
            source (Image or str): The source Image. If the argument is a string, then the source image is loaded from file.
        """
        ...

    def save(self, filename: str):
        """
        Saves this image as a .png file.

        Args:
            filename (str): The path to the file to be saved.

        Raises:
            TypeError: filename is not a string.
            OSError: There was a problem saving the file.
        """
        ...

empty = Image('empty')


"""Support for MP3-compressed audio files

For more information about working with MP3 files in CircuitPython,
see `this CircuitPython Essentials Learn guide page
<https://learn.adafruit.com/circuitpython-essentials/circuitpython-mp3-audio>`_.
"""

from __future__ import annotations

import typing
from typing import Union

from circuitpython_typing import WriteableBuffer

class MP3Decoder:
    """Load a mp3 file for audio playback

    .. note::

        ``MP3Decoder`` uses a lot of contiguous memory, so care should be given to
        optimizing memory usage.  More information and recommendations can be found here:
        https://learn.adafruit.com/Memory-saving-tips-for-CircuitPython/reducing-memory-fragmentation
    """

    def __init__(
        self, file: Union[str, typing.BinaryIO], buffer: WriteableBuffer
    ) -> None:
        """Load a .mp3 file for playback with `audioio.AudioOut` or `audiobusio.I2SOut`.

        :param Union[str, typing.BinaryIO] file: The name of a mp3 file (preferred) or an already opened mp3 file
        :param ~circuitpython_typing.WriteableBuffer buffer: Optional pre-allocated buffer, that will be split in half and used for double-buffering of the data. If not provided, two buffers are allocated internally.  The specific buffer size required depends on the mp3 file.

        Playback of mp3 audio is CPU intensive, and the
        exact limit depends on many factors such as the particular
        microcontroller, SD card or flash performance, and other
        code in use such as displayio. If playback is garbled,
        skips, or plays as static, first try using a "simpler" mp3:

          * Use constant bit rate (CBR) not VBR or ABR (variable or average bit rate) when encoding your mp3 file
          * Use a lower sample rate (e.g., 11.025kHz instead of 48kHz)
          * Use a lower bit rate (e.g., 32kbit/s instead of 256kbit/s)

        Reduce activity taking place at the same time as
        mp3 playback. For instance, only update small portions of a
        displayio screen if audio is playing. Disable auto-refresh
        and explicitly call refresh.

        Playing a mp3 file from flash::

          import board
          import audiomp3
          import audioio
          import digitalio

          # Required for CircuitPlayground Express
          speaker_enable = digitalio.DigitalInOut(board.SPEAKER_ENABLE)
          speaker_enable.switch_to_output(value=True)

          mp3 = audiomp3.MP3Decoder("cplay-16bit-16khz-64kbps.mp3")
          a = audioio.AudioOut(board.A0)

          print("playing")
          a.play(mp3)
          while a.playing:
            pass
          print("stopped")
        """
        ...
    def deinit(self) -> None:
        """Deinitialises the MP3 and releases all memory resources for reuse."""
        ...
    def __enter__(self) -> MP3Decoder:
        """No-op used by Context Managers."""
        ...
    def __exit__(self) -> None:
        """Automatically deinitializes the hardware when exiting a context. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...
    file: typing.BinaryIO
    """File to play back."""
    def open(self, filepath: str) -> None:
        """Takes in the name of a mp3 file, opens it, and replaces the old playback file."""
        ...
    sample_rate: int
    """32 bit value that dictates how quickly samples are loaded into the DAC
    in Hertz (cycles per second). When the sample is looped, this can change
    the pitch output without changing the underlying sample."""
    bits_per_sample: int
    """Bits per sample. (read only)"""
    channel_count: int
    """Number of audio channels. (read only)"""
    rms_level: float
    """The RMS audio level of a recently played moment of audio. (read only)"""
    samples_decoded: int
    """The number of audio samples decoded from the current file. (read only)"""



from turcar.plugins.micropython import (
    BareMetalMicroPythonConfigPage,
    BareMetalMicroPythonProxy,
    add_micropython_backend,
)
from turcar.plugins.micropython.mp_front import VIDS_PIDS_TO_AVOID_IN_GENERIC_BACKEND

LEGO_INVENTOR_PRIME_VIDS_PIDS = {
    (0x0694, 0x0009),  # Spike Prime
    (0x0694, 0x000D),  # Spike Essential
    (0x0694, 0x0010),  # Robot Inventor
}


class PrimeInventorMicroPythonProxy(BareMetalMicroPythonProxy):
    @classmethod
    def should_consider_unknown_devices(cls):
        return False

    @classmethod
    def get_known_usb_vids_pids(cls):
        # https://github.com/pybricks/technical-info/blob/master/assigned-numbers.md
        # can be anything
        return LEGO_INVENTOR_PRIME_VIDS_PIDS

    def get_node_label(self):
        return "Robot Inventor / Spike Prime"

    def _get_backend_launcher_path(self) -> str:
        import turcar.plugins.prime_inventor.prime_inventor_back

        return turcar.plugins.prime_inventor.prime_inventor_back.__file__


class PrimeInventorMicroPythonConfigPage(BareMetalMicroPythonConfigPage):
    def _get_intro_text(self):
        return "Instructions:"

    def _get_intro_url(self):
        return "https://github.com/turcar/turcar/wiki/InventorPrime"

    def may_have_rtc(self):
        return False

    @property
    def allow_webrepl(self):
        return False


def _load_plugin():
    add_micropython_backend(
        "PrimeInventorMicroPython",
        PrimeInventorMicroPythonProxy,
        "MicroPython (Robot Inventor / Spike Prime)",
        PrimeInventorMicroPythonConfigPage,
        bare_metal=True,
        sort_key="24",
        validate_time=False,
        sync_time=False,
    )
    VIDS_PIDS_TO_AVOID_IN_GENERIC_BACKEND.update(LEGO_INVENTOR_PRIME_VIDS_PIDS)



"""Board specific pin names

Common container for board base pin names. These will vary from board to
board so don't expect portability when using this module.

Another common use of this module is to use serial communication buses with
the default pins and settings.  For more information about serial communcication
in CircuitPython, see the :mod:`busio`.

For more information regarding the typical usage of :py:mod:`board`, refer to the `CircuitPython
Essentials Learn guide
<https://learn.adafruit.com/circuitpython-essentials/circuitpython-pins-and-modules>`_

.. warning:: The board module varies by board. The APIs documented here may or may not be
             available on a specific board."""

from __future__ import annotations

import busio

board_id: str
"""Board ID string. The unique identifier for the board model in
circuitpython, as well as on circuitpython.org.
Example: "hallowing_m0_express"."""

def I2C() -> busio.I2C:
    """Returns the `busio.I2C` object for the board's designated I2C bus(es).
    The object created is a singleton, and uses the default parameter values for `busio.I2C`.
    """
    ...

def SPI() -> busio.SPI:
    """Returns the `busio.SPI` object for the board's designated SPI bus(es).
    The object created is a singleton, and uses the default parameter values for `busio.SPI`.
    """
    ...

def UART() -> busio.UART:
    """Returns the `busio.UART` object for the board's designated UART bus(es).
    The object created is a singleton, and uses the default parameter values for `busio.UART`.
    """
    ...
